# We use a hack to get access to the calling routine in the Channel
# object to determine which user operates the channel.  This is done
# by using the behaviour of caller which, if called from within the DB
# package, returns the subroutine arguments of the call frame returned
# in the DB::args array.  This behaviour is documented in the perlfunc
# man page.  You may consider this usage distasteful.

package DB;
use Carp;

@DB::args = ();

# callingClient returns the Client reference of the client object which
# called the first Channel subroutine in the call chain.  This routine
# may only be called from within a Channel object.

# If the calling subroutine is not called from a Client, undef is returned.

# This is a special purpose routine which is useful only within the
# Channel package.

sub _callingClient {
    my $self = shift;

    if (not $self->isa('Channel')) {
	confess "not called as a Channel subroutine";
    }

    my ($package, $filename, $line, $subroutine,
	$hasargs, $wantarray, $evaltext, $is_require);

    # Find first stack frame which does not have a Channel as first
    # argument.  The first argument is considered to be the $self
    # reference.

    my $level = 1;
    while (($package, $filename, $line, $subroutine,
	    $hasargs, $wantarray, $evaltext, $is_require) = caller($level)) {

	if (not $hasargs
	    or not ref($DB::args[0])) {
	    confess "caller not in object context";
	}
	last if (ref($DB::args[0]) ne 'Channel');
	$level++;
    }

    if ($DB::args[0]->isa("Client")) {
	return $DB::args[0];
    } else {
	confess "caller ($DB::args[0]) is not a Client";
    }
}

package Channel;
use Exporter;
use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);

use Carp;

use strict;

my %Extent;

sub callingClient {
    my $self = shift;

    return DB::_callingClient($self);
}

sub findOrCreate {
    my $number = shift;

    if (defined $Extent{lc($number)}) {
	return $Extent{lc($number)};
    } else {
	return $Extent{lc($number)} = new Channel($number);
    }
}

sub defined {
    return defined $Extent{lc(shift)};
}

sub new {
    my $class = shift;
    my $number = shift;
    my $self = {};

    bless $self, $class;

    $self->{'number'} = $number;

    $self->{'kickWords'} = {};
    $self->{'members'} = {};
    $self->{'topic'} = '';

    $self;
}

sub ignoredWord
{
    my $self = shift;
    my $ignoredWord = shift;

    if (not defined $ignoredWord
	or $ignoredWord eq "") {
	$self->{'ignoredWords'} = {}
    } else {
	${$self->{'ignoredWords'}}{lc($ignoredWord)} = 1;
    }
}    

sub isIgnoredWord
{
    my $self = shift;
    my $ignoredWord = shift;

    return defined ${$self->{'ignoredWords'}}{lc($ignoredWord)};
}

sub badWord
{
    my $self = shift;
    my $badWord = shift;

    if (not defined $badWord
	or $badWord eq "") {
	$self->{'kickWords'} = {}
    } else {
	${$self->{'kickWords'}}{lc($badWord)} = 1;
    }
}

sub isBadWord
{
    my $self = shift;
    my $word = shift;

    return defined ${$self->{'kickWords'}}{lc($word)};
}

sub topic
{
    my $self = shift;
    my $newTopic = shift;

    if (defined $newTopic) {
	$self->{'topic'} = $newTopic;
	$self->send("114 " . $self->callingClient->nick() 
		    . " changed the channel topic to '"
		    . $newTopic . "'");
    }

    return $self->{'topic'};
}

sub join
{
    my $self = shift;
    my $client = shift;

    if (not defined $client) {
	$client = $self->callingClient;
    }

    return if (defined $self->{'members'}->{$client});

    $self->{'members'}->{$client} = $client;
    $self->send("232 " . $client->nick . " joined channel " . $self->name);

    return $self;
}

sub leave
{
    my $self = shift;
    my $client = $self->callingClient;

    delete $self->{'members'}->{$client};
    $self->send("231 " . $client->nick . " left channel " . $self->name);
}

sub members
{
    my $self = shift;

    return values %{$self->{'members'}};
}

sub send
{
    my $self = shift;
    my $message = shift;
    my $exclude = shift;
    my $client;
    my $sentTo = 0;

    foreach (split(/\s+/, $message)) {
	if ($self->isIgnoredWord($_)) {
	    return 0;
	}
    }

    foreach $client ($self->members) {
	next if (defined $exclude and $client eq $exclude);
	
	$client->send($message);
	$sentTo++;
    }

    return $sentTo;
}
    
sub number
{
    my $self = shift;

    return $self->{'number'};
}

sub name
{
    my $self = shift;

    return $self->{'number'}; # for now
}

sub isInvisible
{
    my $self = shift;

    return $self->number < 0;
}

1;
