
package Client;
use Net::Ident;
use Exporter;
use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);

use strict;

use POSIX qw(strftime);
use Fcntl;
use Socket;

use conf;

my %Extent;
my $clientCounter = 0;

sub extent
{
    return values %Extent;
}

sub find {
    my $nick = shift;

    foreach (keys %Extent) {
	if (lc($_) eq lc($nick)) {
	    return $Extent{$_};
	}
    }

    return undef;
}

sub fileDescriptor {
    my $self = shift;

    return $self->{'fileDescriptor'};
}

sub all {
    return values %Extent;
}

sub deltaTimeString {
    my $idle = shift;
    if ($idle < 60) {
	return "${idle}s";
    }
    $idle = int($idle/60);
    if ($idle < 60) {
	return "${idle}m";
    }
    $idle = int($idle/60);
    if ($idle < 24) {
	return "${idle}h";
    }
    $idle = int($idle/24);
    return "${idle}d";
}

sub timeStamp {
    return strftime("%d.%m.%y %H:%M:%S ", localtime time);
}

sub logMessage {
    my $line;
    foreach $line (split(/\n/, join("\n", @_))) {
	if ($line ne "") {
	    print STDERR timeStamp(), $line, "\n";
	}
    }
}

sub new {
    my $class = shift;
    my $fileDescriptor = shift;
    my $self = {};

    bless $self, $class;

    fcntl($fileDescriptor, &F_SETFL, &O_NONBLOCK) 
	or die "$0: can't set O_NONBLOCK flag for client socket: $!\n";

    $self->{'ident'} = Net::Ident::lookup($fileDescriptor, 3);
    $self->{'ident'} = "";

    my $sockaddr = getpeername($fileDescriptor);
    if (not defined $sockaddr) {
	warn "$0: could not getpeername for new client - connection dropped\n";
	close($fileDescriptor);
	return undef;		# defend against strange error condition
    }
    my ($port, $addr) = unpack_sockaddr_in($sockaddr);
    my $hostname = gethostbyaddr($addr, AF_INET);
    if (defined $hostname) {
	$self->{'via'} = $hostname;
    } else {
	$self->{'via'} = inet_ntoa($addr);
    }
    if ($self->{'ident'} ne "") {
	$self->{'via'} = $self->{'ident'} . '@' . $self->{'via'};
    }
    $self->{'from'} = "unknown";
    $self->{'fileDescriptor'} = $fileDescriptor;
    $self->{'fileDescriptor'}->autoflush(1);
    $self->{'nick'} = 'unknown' . $clientCounter++;
    $self->{'loggedIn'} = 0;
    $self->{'lastMessageTime'} = time;
    $self->{'loginTime'} = time;
    $self->{'bytesIn'} = $self->{'bytesOut'} = 0;
    $self->{'ignore'} = [];
    $self->{'userinfo'} = '';

    $Extent{$self->{'nick'}} = $self;

    $self->send("201 chat protocol version 1.0\r
100 Welcome to the chat.  Please set your nickname.  Type '.h' for help");

    $self->showFile("$etcDir/motd");

    return $self;
}

sub channel
{
    my $self = shift;

    return $self->{'channel'};
}

sub nick
{
    my $self = shift;

    return $self->{'nick'};
}

sub from
{
    my $self = shift;
    my $newFrom = shift;

    if (defined $newFrom) {
	$self->{'from'} = $newFrom;
    }

    return $self->{'from'};
}

sub addIgnore
{
    my $self = shift;
    my $newIgnore = shift;

    if (not defined $newIgnore
	or $newIgnore eq "") {
	$self->{'ignore'} = [];
    } else {
	unshift @{$self->{'ignore'}}, $newIgnore;
    }
}

sub ignore
{
    my $self = shift;
    my $message = shift;

    foreach (@{$self->{'ignore'}}) {
	if (eval '$message =~ /$_/i') {
	    return 1;
	}
    }

    return 0;
}

sub userinfo
{
    my $self = shift;
    my $arg = shift;

    if (defined $arg) {
	$self->{'userinfo'} = $arg;
    }

    return $self->{'userinfo'};
}

sub send
{
    my $self = shift;
    my $message = shift;

    $self->{'bytesOut'} += length($message);

    $message .= "\r\n";

    return if ($self->ignore($message));

    my $fh = $self->{'fileDescriptor'};

    if (defined $self->{'fileDescriptor'}) {
	if (syswrite($fh, $message, length($message)) != length($message)) {
	    logMessage "$self->{'nick'} bad write ($!)\n";
	    $self->{'fileDescriptor'} = undef;
	    $self->handleCommand(".x");
	}
	if ($debug > 1) {
	    logMessage "$self->{'nick'} MSG: $message";
	} elsif ($debug) {
	    if ($message =~ /^\d{3} /) {
		logMessage "$self->{'nick'} MSG: $message";
	    }
	}
    }
}

sub showFile
{
    my $self = shift;
    my $file = shift;

    if (open(FILE, $file)) {
	while (<FILE>) {
	    chomp;
	    my $cont = "-";
	    if (eof FILE) {
		$cont = " ";
	    }
	    $self->send("100$cont$_");
	}
	close(FILE);
    }
}

sub handleCommand
{
    my $self = shift;
    my $command = shift;

    # take note of the current time (for idle calculations)

    $self->{'lastMessageTime'} = time;
    $self->{'bytesIn'} += length($command);

    # check message length, truncate if oversized.

    if (length($command) > $maxMessageLength) {
	$self->send("301 Message too long, truncated");
	$command = substr($command, 0, $maxMessageLength);
    }

    # log the message -- message text is logged only if $debug is larger
    # than 1

    if ($debug > 1) {
	logMessage "$self->{'nick'} CMD: $command\n";
    } elsif ($debug) {
	my $logText = $command;
	if ($command =~ /^\.m (\S+) /) {
	    $logText = ".m $1 (talk)";
	} elsif ($command !~ /^\./) {
	    $logText = "(talk)";
	}
	logMessage "$self->{'nick'} CMD: $logText\n";
    }

    $command =~ s/\s*$//;		# remove trailing white space
    return if ($command =~ /^\s*$/);	# ignore empty lines.


    ####################################################################
    # message parsing starts here
    ####################################################################

    # first, check for commands which are allowed without logging in

    if ($command =~ /^\.h/) {

	$self->cmdHelp();

	return;

    } elsif ($command =~ /^\.x *(.*)/) {

	$self->cmdExit($1);

	return;

    } elsif ($command =~ /^\.p *(.*)/) {

	$self->cmdShowProtocol($1);

	return;

    } elsif ($command =~ /^\.l\s+(\S+)\s+(\S+)\s*(-?\d{0,5})/) {

	my ($nick, $from, $channel) = ($1, $2, $3);

	$self->cmdLogin($nick, $from, $channel);
	
	return;
    }

    # all other commands require the client to first log in.

    if (not $self->{'loggedIn'}) {
	$self->send("403 Please log in. Type .h for help");
	return;
    }

    # check for remaining commands

    if ($command =~ /^\.n *(\S+)/) {
	
	$self->cmdChangeNick($1);

    } elsif ($command =~ /^\.f *(.*)/) {

	$self->cmdChangeFrom($1);

    } elsif ($command =~ /^\.t *(.*)/) {

	$self->cmdSetChannelTopic($1);

    } elsif ($command =~ /^\.m *(\S+) (.+)$/) {

	my ($to, $message) = ($1, $2);

	$self->cmdSendPrivateMessage($to, $message);

    } elsif ($command =~ /^\.s *(\S*)/) {

	$self->cmdShowClientOrChannel($1);

    } elsif ($command =~ /^\.w$/) {

	$self->cmdWho();

    } elsif ($command =~ /^\.c *(-?\d{0,4})$/i) {

	$self->cmdJoinChannel($1);

    } elsif ($command =~ /^\.i *(.*)/) {

	$self->cmdIgnore($1);

    } elsif ($command =~ /^\.u *(.+)/) {

	$self->userinfo($1);

    } elsif ($command =~ /^\.a *(.+)/) {

	my $action = $1;

	if (not defined $self->channel) {
	    $self->send("100 Not joined to a channel");
	} else {
	    my $sentTo
		= $self->channel->send("118 ".$self->nick." ".$action,
				       $self);
		
	    if ($sentTo == 0) {
		$self->send("100 Nobody sees what you do");
	    }
	}

    } elsif ($command =~ /^\./ 
	     and $command !~ /^\.\../) {

	$self->send("402 Bad command");

    } else {

	if (not defined $self->channel) {

	    $self->send("100 Not joined to a channel");

	} else {

	    my $sentTo
		= $self ->channel->send("<".$self->nick."> ".$command,
					$self);

	    $self->checkDecency($command);

	    if ($sentTo == 0) {
		$self->send("100 Nobody hears you");
	    }
	}
    }
}

sub checkDecency {
    my $self = shift;
    my $talk = shift;

    foreach (split(/\s+/, $talk)) {
	if ($self->{'channel'}->isBadWord($_)) {
	    $self->handleCommand('.x indecent word');
	    return 0;
	}
    }
    return 1;
}

sub readClientData
{
    my $self = shift;
    my $fh = $self->{'fileDescriptor'};
    my $buf;

    my $bytesRead = sysread($fh, $buf, 1024);
    if ($bytesRead <= 0) {
	if ($bytesRead < 0) {
	    logMessage "$self->{'nick'} bad read ($!)\n";
	}
	$self->handleCommand(".x");
	return;
    }

    if ($debug > 3) {
	logMessage "read ", length($buf), " bytes\n";
    }

    if (defined $self->{'buf'}) {
	$buf = $self->{'buf'} . $buf;
    }

    while ($buf =~ s/^([^\r\n]+)\r?\n//) {
	my $command = $1;

	$command =~ s/[\000-\037]/?/g;		# remove control chars

	$command =~ s/\377[\360-\372]//g;	# remove telnet options
	$command =~ s/\377[\373-\376].//g;

	$self->handleCommand($command);
    }

    if (length($buf) > 0) {
	$self->{'buf'} = $buf;
    } else {
	$self->{'buf'} = undef;
    }
}

sub stats {
    my $self = shift;
    my $recipient = shift;

    $recipient->send("117 $self->{'nick'} stats sent "
		     . $self->{'bytesIn'} . " received "
		     . $self->{'bytesOut'} . " online "
		     . deltaTimeString(time - $self->{'loginTime'})
		     . " idle "
		     . deltaTimeString(time - $self->{'lastMessageTime'}));
}


sub cmdChangeNick {
    my $self = shift;
    my $nick = shift;
    my $quiet = shift or 0;

    $nick = substr($nick, 0, $maxNickLength);
    if ($nick !~ /^[a-z0-9][a-z0-9._-]*$/i) {

	$self->send("401 Bad characters in nick name");
	return 0;

    }

    if (find($nick)) {
	$self->send("411 Nickname is already in use");
	return 0;
    }

    my $oldNick = $self->{'nick'};
    $self->{'nick'} = $nick;
    $Extent{$nick} = $self;
    delete $Extent{$oldNick};

    if (not $quiet) {
	Chatserver::send("241 $oldNick $nick - nick changed");
    }

    return 1;
}

sub cmdShowProtocol {
    my $self = shift;

    
    my $arg = $1;
    
    if ($arg eq "") {
	$arg = $protoDefaultLines;
    }

    my $logRef = Chatserver::inCoreLog();
    if ($arg =~ /^\d+$/) {
	my $count = $arg;

	my $index = $#{$logRef} - $count + 1;
	if ($index < 0) {
	    $index = 0;
	}
	
	while ($index <= $#{$logRef}) {
	    $self->send($logRef->[$index++]);
	}
    } else {
	my $regexp = $arg;
	my $index = $#{$logRef};
	my @result;
	
	while ($index > -1
	       and $#result < ($protoDefaultLines-1)) {
	    if (eval '$logRef->[$index] =~ m/$regexp/i') {
		unshift @result, $logRef->[$index];
	    }
	    --$index;
	}

	foreach (@result) {
	    $self->send($_);
	}
    }
}

sub cmdShowClientOrChannel {
    my $self = shift;
    my $arg = shift;

    if ($arg eq "*" or $arg eq "" or not defined $arg) {
	$arg = $self->channel->number;
    }

    if ($arg =~ /^(-?\d+)$/) {

	my $showOnlyChannel = ($1 eq "") ? undef : $1;

	if (defined $showOnlyChannel) {
	    if (not Channel::defined($showOnlyChannel)) {
		$self->send("413 Channel $1 is not active");
		return;
	    }
	    $showOnlyChannel = Channel::findOrCreate($showOnlyChannel);
	}

	if (not defined($self->channel)
	    or (defined $showOnlyChannel
		and $showOnlyChannel->isInvisible
		and $showOnlyChannel ne $self->channel
		and $self !~ /^Manager/)) {
	    $self->send("414 Access violation");
	    return;
	}
		     
	$self->send("100 ".
		    sprintf("%5s %-${maxNickLength}s S %-${maxHostLen}s %s",
			    'Chan', 'Nick', 'From', 'Idle'));
	my $nClients = 0;
	my $nVisible = 0;

	foreach (
		 sort { 
		     not defined $Extent{$a}->{'channel'}
		     or not defined $Extent{$b}->{'channel'}
		     or (($Extent{$a}->{'channel'}->number 
			  cmp $Extent{$b}->{'channel'}->number))
			 or (lc($Extent{$a}->{'nick'})
			     cmp lc($Extent{$b}->{'nick'}))
			 }
		 keys %Extent) {

	    my $client = $Extent{$_};
	    $nClients++;

	next if (not defined $client->{'channel'});

	    next if ($self !~ /^Manager/
		     and $client->{'channel'}->isInvisible
		     and $client->{'channel'} ne $self->{'channel'});

	    next if (defined $showOnlyChannel
		     and ($client->channel ne $showOnlyChannel));
	    
	    if ($client->{'loggedIn'}) {
		$nVisible++;
		my $status = "R";
		my $host = $client->{'via'};
		if ($host =~ /localhost$/) {
		    $host = $client->from;
		    $status = "L";
		}
		if (length($host) > $maxHostLen) {
		    $host = substr($host, 0, $maxHostLen-1) . "*";
		}
		my $idle = deltaTimeString(time - $client->{'lastMessageTime'});
		$self->send("110 ".
			    sprintf("%5d %-$maxNickLength.${maxNickLength}s %s %-$maxHostLen.${maxHostLen}s %-3s",
				    $client->{'channel'}->number, 
				    $client->{'nick'},
				    $status, $host, $idle));
	    }
	}

	if (defined $arg) {	# we know that it is a valid channel
	    my $channel = Channel::findOrCreate($arg);
	    $self->send("115 $arg ".$channel->topic);
	}

	$self->send(sprintf("100 %d/%d user%s logged on/invisible",
			    $nClients, $nClients-$nVisible,
			    (($nClients != 1) ? "s" : "")));
    } else {
	my $client = find($arg);

	if (not defined $client) {
	    $self->send("412 bad nick name $client");
	} else {
	    $self->send("116 "
			. $client->nick . " "
			. $client->from . " "
			. $client->{'via'} . " "
			. $client->userinfo);

	    $client->stats($self);
	}
    }
}

sub cmdSendPrivateMessage {
    my $self = shift;
    my ($to, $message) = @_;

    my $sent = 0;

    foreach (keys %Extent) {

	if (lc($_) eq lc($to)) {
	    $Extent{$_}->send("*$self->{'nick'}* $message");
	    $sent = 1;
	    last;
	}
    }

    if (not $sent) {
	$self->send("412 Bad nickname, message not sent");
    }

}

sub cmdWho {
    my $self = shift;


    my @lines = `/usr/bin/w`;
    while ($#lines > 0) {
	$_ = shift @lines;
	chomp;
	$self->send("112-$_") if (/^vchat/ or /^USER/ or /load aver/);
    }
    $_ = ((shift @lines) or "no output created by sub process.");
    chomp;
    $self->send("112 $_");
}

sub cmdJoinChannel {
    my $self = shift;
    my $channel = shift;

    $self->{'channel'}->leave();
    $self->{'channel'} = Channel::findOrCreate($channel)->join();
}

sub cmdIgnore {
    my $self = shift;
    my $regexp = shift;

    $self->addIgnore($regexp);
    $self->send("113 ignorance increased");

}

sub cmdSetChannelTopic {
    my $self = shift;
    my $topic = shift;

    if (length($topic) > $maxTopicLen) {
	$self->send("302 Topic too long, maximum is $maxTopicLen characters");
    } elsif ($topic eq "") {
	$self->send("115 " . $self->channel->name . " " 
		    . $self->channel->topic);
    } else {
	$self->channel->topic($topic);
    }
}

sub cmdLogin {
    my $self = shift;
    my ($nick, $from, $channel) = @_;

    if ($self->{'loggedIn'}) {
	$self->send("404 Already logged in");
	return;
    }

    if (find($nick)) {
	$self->send("411 Nickname is already in use");
	return;
    }

    if (not $self->cmdChangeNick($nick, 1)) {
	return;
    }

    $self->{'loggedIn'} = 1;
    $self->from($from);

  Chatserver::send("211 $self->{'nick'} ($self->{'from'}/$self->{'via'}) entered the chat");

    if ($channel eq "") {
	$channel = 0;
    }
    $self->{'channel'} = Channel::findOrCreate($channel)->join();

}

sub cmdHelp {
    my $self = shift;


	$self->send(
"100-This is a multi-party chat system.  People who are on one channel\r
100-see all text entered by other people on that channel.\r
100-\r
100-Commands:\r
100-.s [<channel>|<nick>] Show status information about channels/users\r
100-.w                    Show users logged into the server\r
100-.p [<count>|<regexp>] Show server log file (<count> limits the number of\r
100-                      lines, <regexp> searches the in-core log)\r
100-.c <chan>             Change to channel with number <chan>\r
100-.n <nick>             Change nickname to <nick>\r
100-.m <nick> <msg>       Send private message to <nick>\r
100-.a <something>        Do <something>\r
100-.l <nick> <from>      Log in (performed automatically by client)\r
100-.i <regexp>           Add <regexp> to your ignorance list\r
100-.u <text>             Set up user information (can be shown with .s <nick>)\r
100-.t [<topic>]          Show/set channel topic\r
100-.x <reason>           Exit\r
100-.h                    You just found out what this does.\r
100-\r
100 Problem reports may be directed to hans\@berlin.ccc.de");

}

sub cmdExit {
    my $self = shift;
    my $reason = shift;


    return if ($self->{'leaving'});

    $self->{'leaving'} = 1;

    if (defined $self->channel) {
	$self->channel->leave;
    }

    if ($self->{'loggedIn'}) {
      Chatserver::send("221 $self->{'nick'} left the chat"
		       . ($reason ne "" ? " ($reason)" : ""));
	$self->stats(Chatserver::instance());
    }
    if (defined $self->{'fileDescriptor'}) {
	close($self->{'fileDescriptor'});
    }
    delete $Extent{$self->{'nick'}};
}

sub cmdChangeFrom {
    my $self = shift;
    my $newFrom = shift;

    if (length($newFrom) > $maxHostLen) {
	$self->send("302 From too long");
	return;
    }

    if ($newFrom =~ /\s/) {
	$self->send("401 Bad characters in from string");
	return;
    }

    $self->from($newFrom);
}

1;
