#!/usr/local/bin/perl5

# chatserver.pl - Multi-Party chat server.

# This is a TCP/IP based chat server.  To participate in the chat,
# clients connect to this server using a TCP/IP connection.  Messages
# received from a client are sent out to all other clients who are
# logged to the same channel.  Special messages which are meant to be
# interpreted by the server rather than distributed to the channel
# participants are begin with a dot ('.').

# The protocol is kept human readable so that people can participate
# without having to use special client software simply by using telnet
# to the server port.

# Clients need to first log in using the '.l' command.

# A second server socked is opened for manager log ins.  This port
# is meant to be protected by the firewall configuration.  Users who
# logged in through the manager port have enhanced privileges, see
# Manager::handleCommand.

# $Header: /home/cvsroot/vchat/server/chatserver.pl,v 1.18 1998/08/15 19:42:31 hans Exp $

# $Log: chatserver.pl,v $
# Revision 1.18  1998/08/15 19:42:31  hans
# Suppress statistics.
#
# Revision 1.17  1998/07/27 23:12:17  hans
# Make Chatserver a singleton with a send subroutine to unify broadcast
# handling.
#
# Revision 1.16  1998/07/27 22:43:24  hans
# Defend against clients dying during the processing of a message in
# other channels during one select run.
#
# Revision 1.15  1998/07/26 19:27:47  hans
# Guard against accept() returning undef.
# Rearrange for emacs perl mode sanity.
# Suppress 'bad read' messages.
#
# Revision 1.14  1998/07/04 16:41:48  hans
# Fix up logging.  Only broadcast messages are written to the log file.
# Fix .t - If used without an argument, the current channel topic is shown.
# Allow for a channel other than 0 to be joined when logging in with .l.
#
# Revision 1.13  1998/07/04 00:03:22  hans
# Moved more functionality into the Channel object.  Messages are now
# generally sent via the send() subroutine, which is implemented by both
# Channel and Client objects.  Logging is performed in the new
# Chatserver class, which is implemented in chatserver.pl.  Joining and
# leaving a channel is performed by the Channel class.  Channel topics
# are partially implemented (.t command).
#
# Revision 1.12  1998/05/02 13:03:11  hans
# Completely reworked version.
# Modularized Client/Channel/Manager
#
# Revision 1.11  1998/04/23 09:06:45  hans
# Negative Kanalnummern nicht im Protokoll mitloggen.
#
# Revision 1.10  1998/04/21 11:42:02  hans
# Mithoeren ist nicht mehr moeglich, wenn man nicht eingelogged ist.
#
# Revision 1.9  1998/04/14 00:33:21  hans
# idle field for clients added.
# show only one host name (localhost=>FROM specification, other=>peer name)
# negative channels added, will not be shown in .s
#
# Revision 1.8  1998/04/13 23:13:05  hans
# pass lines beginning with .. to the channel. (djenia)
#
# Revision 1.7  1998/04/13 22:05:31  hans
# Added system paths
# Print $etcDir/motd to clients connecting.
#
# Revision 1.6  1998/04/13 21:40:39  hans
# moving to cvs
#
# Revision 1.5  1998/03/09 11:52:24  hans
# Cosmetic changes.
#
# Revision 1.4  1998/03/07 21:58:24  hans
# Manager port added.
# Structural changes.
# Minor fixes.
#
# Revision 1.3  1998/03/07 19:45:25  hans
# Remove telnet options.
# Fix 'protocol' command.
#

package Chatserver;

use English;
use strict;
use FileHandle;
use IO::Socket;
use Net::hostent;
use POSIX qw(strftime);

use Channel;
use Manager;
use Client;

use conf;

my $instance = {};
bless $instance, "Chatserver";

sub instance
{
    return $instance;
}

sub send
{
    my $message = shift;

    if (ref($message)) {
	$message = shift;
    }
    my $client;

    logEnt($message);

    # Broadcast the message to all Clients.  Statistics are suppressed
    if ($message !~ /^117/) {
	foreach $client (Client::extent) {
	    $client->send($message);
	}
    }
}

my @PROTOCOL;

sub timeStamp {
    return strftime("%d.%m.%y %H:%M:%S ", localtime time);
}

sub logEnt
{
    my $message = shift;

    my $logEnt = "111 " . timeStamp() . $message;
    print $PROTOCOL $logEnt, "\n";
    push @PROTOCOL, $logEnt;

    if ($#PROTOCOL >= $maxProtocol) {
	unshift @PROTOCOL;
    }
}

sub inCoreLog
{
    return \@PROTOCOL;
}

my $clientSocket = IO::Socket::INET->new( Proto     => 'tcp',
					  LocalPort => $clientPort,
					  Listen    => &SOMAXCONN,
					  Reuse     => 1);

$clientSocket or die "$0: can't setup client listen socket ($!)\n";

my $managerSocket = IO::Socket::INET->new( Proto     => 'tcp',
					   LocalPort => $managerPort,
					   Listen    => &SOMAXCONN,
					   Reuse     => 1);

$managerSocket or die "$0: can't setup manager listen socket ($!)\n";

$SIG{PIPE} = 'IGNORE';

print STDERR "$0: accepting chat clients on port $clientPort\n";

while (1) {

    # then set up select set

    my $ready = '';
    vec($ready, fileno($clientSocket), 1) = 1;
    vec($ready, fileno($managerSocket), 1) = 1;

    foreach (Client::all()) {
	vec($ready, fileno($_->fileDescriptor), 1) = 1;
    }

    my $count = select($ready, undef, undef, undef);

    if (vec($ready, fileno($clientSocket), 1)) {
	my $newClientFD = $clientSocket->accept();
	if (defined $newClientFD) {
	    new Client($newClientFD);
	}
    }

    if (vec($ready, fileno($managerSocket), 1)) {
	new Manager($managerSocket->accept());
    }

    foreach (Client::all()) {
	next if (not defined $_);	# defend against dead clients
				
	if (vec($ready, fileno($_->fileDescriptor), 1)) {
	    $_->readClientData;
	}
    }
}

